#pragma once

#include "../common.h"

namespace doukutsu_rs::engine_constants::npcs
{
    extern const common::Rect<uint16_t> EMPTY_NPC_RECT;

    template <size_t T>
    class SafeNPCRect
    {
    public:
        std::array<common::Rect<uint16_t>, T> rects;

        SafeNPCRect(common::Rect<uint16_t> rects[T])
        {
            for (size_t i = 0; i < T; i++)
            {
                this->rects[i] = rects[i];
            }
        }

        common::Rect<uint16_t> const &operator[](size_t i)
        {
            if (i >= T)
            {
                return EMPTY_NPC_RECT;
            }
            else
            {
                return rects[i];
            }
        }
    };

    struct NPCConsts
    {
        SafeNPCRect<6> n001_experience;
        SafeNPCRect<14> n002_behemoth;
        SafeNPCRect<16> n004_smoke;
        SafeNPCRect<6> n005_green_critter;
        SafeNPCRect<10> n006_green_beetle;
        SafeNPCRect<6> n007_basil;
        SafeNPCRect<4> n008_blue_beetle;
        SafeNPCRect<6> n009_balrog_falling_in;
        SafeNPCRect<8> n010_balrog_shooting;
        SafeNPCRect<3> n011_balrog_energy_shot;
        SafeNPCRect<28> n012_balrog_cutscene;
        SafeNPCRect<4> n013_forcefield;
        SafeNPCRect<3> n014_key;
        SafeNPCRect<3> n015_closed_chest;
        SafeNPCRect<8> n016_save_point;
        SafeNPCRect<2> n017_health_refill;
        SafeNPCRect<2> n018_door;
        SafeNPCRect<8> n019_balrog_bust_in;
        SafeNPCRect<4> n020_computer;
        SafeNPCRect<2> n021_chest_open;
        SafeNPCRect<2> n022_teleporter;
        SafeNPCRect<8> n023_teleporter_lights;
        SafeNPCRect<12> n024_power_critter;
        SafeNPCRect<2> n025_lift;
        SafeNPCRect<8> n026_bat_flying;
        SafeNPCRect<2> n027_death_trap;
        SafeNPCRect<12> n028_flying_critter;
        SafeNPCRect<4> n029_cthulhu;
        SafeNPCRect<3> n030_hermit_gunsmith;
        SafeNPCRect<10> n031_bat_hanging;
        SafeNPCRect<2> n032_life_capsule;
        SafeNPCRect<2> n033_balrog_bouncing_projectile;
        SafeNPCRect<2> n034_bed;
        SafeNPCRect<8> n035_mannan;
        SafeNPCRect<12> n036_balrog_hover;
        SafeNPCRect<2> n037_sign;
        SafeNPCRect<4> n038_fireplace;
        SafeNPCRect<2> n039_save_sign;
        SafeNPCRect<14> n040_santa;
        common::Rect<uint16_t> n041_busted_door;
        SafeNPCRect<26> n042_sue;
        SafeNPCRect<2> n043_chalkboard;
        SafeNPCRect<6> n044_polish;
        SafeNPCRect<3> n045_baby;
        SafeNPCRect<5> n047_sandcroc;
        SafeNPCRect<4> n048_omega_projectiles;
        SafeNPCRect<6> n049_skullhead;
        SafeNPCRect<4> n050_skeleton_projectile;
        SafeNPCRect<10> n051_crow_and_skullhead;
        SafeNPCRect<2> n052_sitting_blue_robot;
        SafeNPCRect<4> n053_skullstep_leg;
        SafeNPCRect<6> n054_skullstep;
        SafeNPCRect<12> n055_kazuma;
        SafeNPCRect<6> n056_tan_beetle;
        SafeNPCRect<10> n057_crow;
        SafeNPCRect<6> n058_basu;
        SafeNPCRect<4> n059_eye_door;
        SafeNPCRect<16> n060_toroko;
        SafeNPCRect<22> n061_king;
        SafeNPCRect<3> n062_kazuma_computer;
        SafeNPCRect<12> n063_toroko_stick;
        SafeNPCRect<6> n064_first_cave_critter;
        SafeNPCRect<8> n065_first_cave_bat;
        SafeNPCRect<4> n066_misery_bubble;
        SafeNPCRect<16> n067_misery_floating;
        SafeNPCRect<18> n068_balrog_running;
        SafeNPCRect<12> n069_pignon;
        SafeNPCRect<4> n070_sparkle;
        SafeNPCRect<6> n071_chinfish;
        SafeNPCRect<2> n072_sprinkler;
        SafeNPCRect<5> n073_water_droplet;
        SafeNPCRect<12> n074_jack;
        SafeNPCRect<2> n075_kanpachi;
        SafeNPCRect<3> n077_yamashita;
        SafeNPCRect<2> n078_pot;
        SafeNPCRect<6> n079_mahin;
        SafeNPCRect<14> n080_gravekeeper;
        SafeNPCRect<12> n081_giant_pignon;
        SafeNPCRect<18> n082_misery_standing;
        SafeNPCRect<16> n083_igor_cutscene;
        SafeNPCRect<4> n084_basu_projectile;
        SafeNPCRect<6> n085_terminal;
        SafeNPCRect<5> n086_missile_pickup;
        SafeNPCRect<5> n087_heart_pickup;
        SafeNPCRect<24> n088_igor_boss;
        SafeNPCRect<8> n089_igor_dead;
        common::Rect<uint16_t> n090_background;
        common::Rect<uint16_t> n091_mimiga_cage;
        SafeNPCRect<3> n092_sue_at_pc;
        SafeNPCRect<14> n093_chaco;
        SafeNPCRect<5> n094_kulala;
        SafeNPCRect<8> n095_jelly;
        SafeNPCRect<3> n096_fan_left;
        SafeNPCRect<3> n097_fan_up;
        SafeNPCRect<3> n098_fan_right;
        SafeNPCRect<3> n099_fan_down;
        SafeNPCRect<2> n100_grate;
        SafeNPCRect<3> n101_malco_screen;
        SafeNPCRect<4> n102_malco_computer_wave;
        SafeNPCRect<6> n103_mannan_projectile;
        SafeNPCRect<6> n104_frog;
        SafeNPCRect<2> n105_hey_bubble_low;
        SafeNPCRect<10> n107_malco_broken;
        SafeNPCRect<3> n108_balfrog_projectile;
        SafeNPCRect<4> n109_malco_powered_on;
        SafeNPCRect<6> n110_puchi;
        SafeNPCRect<4> n111_quote_teleport_out;
        SafeNPCRect<4> n112_quote_teleport_in;
        SafeNPCRect<14> n113_professor_booster;
        SafeNPCRect<3> n114_press;
        SafeNPCRect<12> n115_ravil;
        common::Rect<uint16_t> n116_red_petals;
        SafeNPCRect<20> n117_curly;
        SafeNPCRect<18> n118_curly_boss;
        common::Rect<uint16_t> n119_table_chair;
        SafeNPCRect<2> n120_colon_a;
        SafeNPCRect<3> n121_colon_b;
        SafeNPCRect<20> n122_colon_enraged;
        SafeNPCRect<4> n123_curly_boss_bullet;
        SafeNPCRect<2> n124_sunstone;
        SafeNPCRect<2> n125_hidden_item;
        SafeNPCRect<12> n126_puppy_running;
        SafeNPCRect<6> n127_machine_gun_trail_l2;
        SafeNPCRect<20> n128_machine_gun_trail_l3;
        SafeNPCRect<18> n129_fireball_snake_trail;
        SafeNPCRect<8> n130_puppy_sitting;
        SafeNPCRect<2> n131_puppy_sleeping;
        SafeNPCRect<10> n132_puppy_barking;
        SafeNPCRect<4> n133_jenka;
        SafeNPCRect<6> n134_armadillo;
        SafeNPCRect<4> n135_skeleton;
        SafeNPCRect<4> n136_puppy_carried;
        common::Rect<uint16_t> n137_large_door_frame;
        SafeNPCRect<2> n138_large_door;
        SafeNPCRect<6> n139_doctor;
        SafeNPCRect<28> n140_toroko_frenzied;
        SafeNPCRect<2> n141_toroko_block_projectile;
        SafeNPCRect<5> n142_flower_cub;
        SafeNPCRect<2> n143_jenka_collapsed;
        SafeNPCRect<10> n144_toroko_teleporting_in;
        SafeNPCRect<2> n145_king_sword;
        SafeNPCRect<5> n146_lightning;
        SafeNPCRect<12> n147_critter_purple;
        SafeNPCRect<2> n148_critter_purple_projectile;
        common::Rect<uint16_t> n149_horizontal_moving_block;
        SafeNPCRect<20> n150_quote;
        SafeNPCRect<4> n151_blue_robot_standing;
        SafeNPCRect<14> n153_gaudi;
        SafeNPCRect<6> n154_gaudi_dead;
        SafeNPCRect<8> n155_gaudi_flying;
        SafeNPCRect<3> n156_gaudi_projectile;
        common::Rect<uint16_t> n157_vertical_moving_block;
        SafeNPCRect<8> n158_fish_missile;
        common::Rect<uint16_t> n159_monster_x_defeated;
        SafeNPCRect<8> n160_puu_black;
        SafeNPCRect<3> n161_puu_black_projectile;
        SafeNPCRect<3> n162_puu_black_dead;
        SafeNPCRect<4> n163_dr_gero;
        SafeNPCRect<4> n164_nurse_hasumi;
        SafeNPCRect<3> n165_curly_collapsed;
        SafeNPCRect<2> n166_chaba;
        SafeNPCRect<3> n167_booster_falling;
        common::Rect<uint16_t> n168_boulder;
        SafeNPCRect<18> n169_balrog_shooting_missiles;
        SafeNPCRect<4> n170_balrog_missile;
        SafeNPCRect<4> n171_fire_whirrr;
        SafeNPCRect<3> n172_fire_whirrr_projectile;
        SafeNPCRect<8> n173_gaudi_armored;
        SafeNPCRect<3> n174_gaudi_armored_projectile;
        SafeNPCRect<4> n175_gaudi_egg;
        SafeNPCRect<6> n176_buyo_buyo_base;
        SafeNPCRect<2> n177_buyo_buyo;
        SafeNPCRect<3> n178_core_blade_projectile;
        SafeNPCRect<3> n179_core_wisp_projectile;
        SafeNPCRect<22> n180_curly_ai;
        SafeNPCRect<4> n181_curly_ai_machine_gun;
        SafeNPCRect<4> n182_curly_ai_polar_star;
        SafeNPCRect<2> n183_curly_air_tank_bubble;
        SafeNPCRect<4> n184_shutter;
        common::Rect<uint16_t> n185_small_shutter;
        SafeNPCRect<4> n186_lift_block;
        SafeNPCRect<4> n187_fuzz_core;
        SafeNPCRect<4> n188_fuzz;
        SafeNPCRect<3> n189_homing_flame;
        SafeNPCRect<2> n190_broken_robot;
        SafeNPCRect<4> n192_scooter;
        common::Rect<uint16_t> n193_broken_scooter;
        common::Rect<uint16_t> n194_broken_blue_robot;
        common::Rect<uint16_t> n195_background_grate;
        SafeNPCRect<2> n196_ironhead_wall;
        SafeNPCRect<4> n197_porcupine_fish;
        SafeNPCRect<3> n198_ironhead_projectile;
        SafeNPCRect<5> n199_wind_particles;
        SafeNPCRect<12> n200_zombie_dragon;
        SafeNPCRect<2> n201_zombie_dragon_dead;
        SafeNPCRect<3> n202_zombie_dragon_projectile;
        SafeNPCRect<6> n203_critter_destroyed_egg_corridor;
        SafeNPCRect<2> n204_small_falling_spike;
        SafeNPCRect<2> n205_large_falling_spike;
        SafeNPCRect<3> n206_counter_bomb;
        SafeNPCRect<5> n207_counter_bomb_countdown;
        SafeNPCRect<6> n208_basu_destroyed_egg_corridor;
        SafeNPCRect<4> n209_basu_projectile_destroyed_egg_corridor;
        SafeNPCRect<4> n210_beetle_destroyed_egg_corridor;
        SafeNPCRect<4> n211_small_spikes;
        SafeNPCRect<4> n212_sky_dragon;
        SafeNPCRect<10> n213_night_spirit;
        SafeNPCRect<3> n214_night_spirit_projectile;
        SafeNPCRect<5> n215_sandcroc_outer_wall;
        common::Rect<uint16_t> n216_debug_cat;
        SafeNPCRect<8> n217_itoh;
        SafeNPCRect<2> n218_core_giant_ball;
        SafeNPCRect<4> n220_shovel_brigade;
        SafeNPCRect<12> n221_shovel_brigade_walking;
        common::Rect<uint16_t> n222_prison_bars;
        SafeNPCRect<6> n223_momorin;
        SafeNPCRect<4> n224_chie;
        SafeNPCRect<4> n225_megane;
        SafeNPCRect<7> n226_kanpachi_plantation;
        common::Rect<uint16_t> n227_bucket;
        SafeNPCRect<8> n228_droll;
        SafeNPCRect<2> n229_red_flowers_sprouts;
        SafeNPCRect<2> n230_red_flowers_blooming;
        SafeNPCRect<2> n231_rocket;
        SafeNPCRect<6> n232_orangebell;
        SafeNPCRect<8> n233_orangebell_bat;
        SafeNPCRect<2> n234_red_flowers_picked;
        SafeNPCRect<8> n235_midorin;
        SafeNPCRect<12> n236_gunfish;
        common::Rect<uint16_t> n237_gunfish_projectile;
        SafeNPCRect<3> n238_press_sideways;
        SafeNPCRect<2> n239_cage_bars;
        SafeNPCRect<12> n240_mimiga_jailed;
        SafeNPCRect<6> n241_critter_red;
        SafeNPCRect<8> n242_bat_last_cave;
        common::Rect<uint16_t> n244_lava_drop;
        SafeNPCRect<4> n245_lava_drop_generator;
        SafeNPCRect<3> n246_press_proximity;
        SafeNPCRect<18> n247_misery_boss;
        SafeNPCRect<3> n248_misery_boss_vanishing;
        SafeNPCRect<2> n249_misery_boss_appearing;
        SafeNPCRect<3> n250_misery_boss_lightning_ball;
        SafeNPCRect<2> n251_misery_boss_lightning;
        SafeNPCRect<8> n252_misery_boss_bats;
        SafeNPCRect<2> n253_experience_capsule;
        SafeNPCRect<2> n254_helicopter;
        SafeNPCRect<8> n255_helicopter_blades;
        SafeNPCRect<6> n256_doctor_facing_away;
        SafeNPCRect<3> n257_red_crystal;
        common::Rect<uint16_t> n258_mimiga_sleeping;
        SafeNPCRect<2> n259_curly_unconscious;
        SafeNPCRect<6> n260_shovel_brigade_caged;
        SafeNPCRect<4> n261_chie_caged;
        SafeNPCRect<4> n262_chaco_caged;
        SafeNPCRect<18> n263_doctor_boss;
        common::Rect<uint16_t> n264_doctor_boss_red_projectile;
        SafeNPCRect<3> n265_doctor_boss_red_projectile_trail;
        SafeNPCRect<2> n266_doctor_boss_red_projectile_bouncing;
        SafeNPCRect<20> n267_muscle_doctor;
        SafeNPCRect<20> n268_igor_enemy;
        SafeNPCRect<6> n269_red_bat_bouncing;
        SafeNPCRect<2> n270_doctor_red_energy;
        SafeNPCRect<3> n273_droll_projectile;
        SafeNPCRect<12> n274_droll;
        SafeNPCRect<4> n275_puppy_plantation;
        SafeNPCRect<18> n276_red_demon;
        SafeNPCRect<3> n277_red_demon_projectile;
        SafeNPCRect<6> n278_little_family;
        SafeNPCRect<2> n279_large_falling_block;
        SafeNPCRect<4> n280_sue_teleported;
        SafeNPCRect<3> n282_mini_undead_core_active;
        SafeNPCRect<22> n283_misery_possessed;
        SafeNPCRect<26> n284_sue_possessed;
        common::Rect<uint16_t> n285_undead_core_spiral_projectile;
        SafeNPCRect<3> n286_undead_core_spiral_projectile_trail;
        SafeNPCRect<7> n287_orange_smoke;
        SafeNPCRect<5> n288_undead_core_exploding_rock;
        SafeNPCRect<6> n289_critter_orange;
        SafeNPCRect<6> n290_bat_misery;
        SafeNPCRect<2> n291_mini_undead_core_inactive;
        SafeNPCRect<2> n293_undead_core_energy_shot;
        SafeNPCRect<4> n295_cloud;
        common::Rect<uint16_t> n297_sue_dragon_mouth;
        SafeNPCRect<8> n298_intro_doctor;
        SafeNPCRect<2> n299_intro_balrog_misery;
        common::Rect<uint16_t> n300_intro_demon_crown;
        SafeNPCRect<8> n301_misery_fish_missile;
        SafeNPCRect<4> n303_curly_machine_gun;
        SafeNPCRect<4> n304_gaudi_hospital;
        SafeNPCRect<4> n305_small_puppy;
        SafeNPCRect<4> n306_balrog_nurse;
        SafeNPCRect<4> n307_santa_caged;
        SafeNPCRect<4> n308_stumpy;
        SafeNPCRect<4> n309_bute;
        SafeNPCRect<10> n310_bute_sword;
        SafeNPCRect<14> n311_bute_archer;
        SafeNPCRect<10> n312_bute_arrow_projectile;
        SafeNPCRect<28> n313_ma_pignon;
        SafeNPCRect<3> n314_ma_pignon_rock;
        SafeNPCRect<8> n315_ma_pignon_clone;
        SafeNPCRect<6> n316_bute_dead;
        SafeNPCRect<8> n317_mesa;
        SafeNPCRect<6> n318_mesa_dead;
        SafeNPCRect<3> n319_mesa_block;
        SafeNPCRect<6> n320_curly_carried;
        SafeNPCRect<6> n321_curly_nemesis;
        SafeNPCRect<3> n322_deleet;
        SafeNPCRect<4> n323_bute_spinning;
        SafeNPCRect<7> n325_heavy_press_lightning;
        SafeNPCRect<16> n326_sue_itoh_human_transition;
        SafeNPCRect<2> n327_sneeze;
        common::Rect<uint16_t> n328_human_transform_machine;
        SafeNPCRect<2> n329_laboratory_fan;
        SafeNPCRect<3> n330_rolling;
        SafeNPCRect<4> n331_ballos_bone_projectile;
        SafeNPCRect<3> n332_ballos_shockwave;
        SafeNPCRect<2> n333_ballos_lightning;
        SafeNPCRect<4> n334_sweat;
        SafeNPCRect<3> n335_ikachan;
        SafeNPCRect<2> n337_numahachi;
        SafeNPCRect<4> n338_green_devil;
        SafeNPCRect<22> n340_ballos;
        SafeNPCRect<3> n341_ballos_1_head;
        SafeNPCRect<3> n342_ballos_orbiting_eye;
        common::Rect<uint16_t> n343_ballos_3_cutscene;
        SafeNPCRect<2> n344_ballos_3_eyes;
        SafeNPCRect<4> n345_ballos_skull_projectile;
        common::Rect<uint16_t> n346_ballos_orbiting_platform;
        SafeNPCRect<4> n347_hoppy;
        SafeNPCRect<2> n348_ballos_4_spikes;
        common::Rect<uint16_t> n349_statue;
        SafeNPCRect<14> n350_flying_bute_archer;
        SafeNPCRect<9> n351_statue_shootable;
        SafeNPCRect<28> n352_ending_characters;
        SafeNPCRect<8> n353_bute_sword_flying;
        SafeNPCRect<4> n355_quote_and_curly_on_balrog;
        SafeNPCRect<2> n356_balrog_rescuing;
        common::Rect<uint16_t> n357_puppy_ghost;
        SafeNPCRect<5> n358_misery_credits;
        common::Rect<uint16_t> n360_credits_thank_you;
        SafeNPCRect<10> b01_omega;
        SafeNPCRect<18> b02_balfrog;
        SafeNPCRect<29> b03_monster_x;
        SafeNPCRect<10> b04_core;
        SafeNPCRect<18> b05_ironhead;
        SafeNPCRect<14> b06_sisters;
        SafeNPCRect<15> b07_undead_core;
        SafeNPCRect<6> b08_heavy_press;
        SafeNPCRect<14> b09_ballos;
    };
}